import inspect

from eventbus.utils import get_class_that_defined_method


class SubscriberMethod:

    def __init__(self, method, event_type, priority: int = 0, sticky: bool = False):
        self.method = method
        self.event_type = event_type
        self.priority = priority
        self.sticky = sticky
        self.check_method_definition()

    def __call__(self, *args, **kwargs):
        self.method(*args, **kwargs)

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif isinstance(other, SubscriberMethod):
            # self.check_method_definition()
            # other.check_method_definition()
            return self.method_definition == other.method_definition
        else:
            return False

    def check_method_definition(self):
        class_name = get_class_that_defined_method(self.method).__name__
        method_name = self.method.__name__
        # todo: 消除潜在bug
        event_type = list(inspect.getfullargspec(self.method).annotations.values())[0].__name__
        self.method_definition = f'{class_name}#{method_name}({event_type})'

    def __repr__(self):
        return f'<SubscriberMethod: {self.method_definition}>'



import functools
import inspect


def get_class_that_defined_method(method):
    if isinstance(method, functools.partial):
        return get_class_that_defined_method(method.func)
    if inspect.ismethod(method) or (inspect.isbuiltin(method) and getattr(method, '__self__', None) is not None and getattr(method.__self__, '__class__', None)):
        for cls in inspect.getmro(method.__self__.__class__):
            if method.__name__ in cls.__dict__:
                return cls
        method = getattr(method, '__func__', method)  # fallback to __qualname__ parsing
    if inspect.isfunction(method):
        cls = getattr(inspect.getmodule(method),
                      method.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0],
                      None)
        if isinstance(cls, type):
            return cls
    return getattr(method, '__objclass__', None)  # handle special descriptor objects


import inspect

subscribe_annotation_property = '__subscribe_annotation__'


class SubscribeAnnotation:

    def __init__(self, priority=0, sticky=False):
        self._priority = priority
        self._sticky = sticky

    @property
    def priority(self):
        return self._priority

    @property
    def sticky(self):
        return self._sticky


class Subscribe:
    """订阅装饰器，给被订阅的方法添加元信息

    订阅装饰器将会为方法添加方法元信息，用来在EventBus对类进行注册时使用。
    注意，为了降低可能产生Bug的情况，这里限制仅允许为类方法进行装饰，但并
    非不能对其他可调用类进行装饰。

    :arg priority 优先级，该方法在同一队列内处理事件的优先级
    :arg sticky 粘滞事件处理标志，在该方法被初始化时是否处理对应事件的粘滞事件
    """

    def __init__(self, priority: int = 100, sticky: bool = False):
        self.priority: int = priority
        self.sticky: bool = sticky

    def __call__(self, callable):
        """调用Subscribe装饰器对可调用对象的修改

        该方法将限制被装饰对象为类方法，由于装饰时均不以对象绑定类方法的形式进行调用，故在装饰时需要进行特殊判断。
        :param callable: 被装饰类方法
        :return:
        """
        self._check_method_valid(callable)
        setattr(callable, subscribe_annotation_property, SubscribeAnnotation(priority=self.priority, sticky=self.sticky))
        return callable

    @staticmethod
    def _check_method_valid(callable):
        # 判断方法签名是否为Class.Method(self, event: EventType)
        # 注意此处不检查classmethod/staticmethod/instancemethod,正确使用时应使用instancemethod
        qualname_splited = callable.__qualname__.split('.')
        arg_spec = inspect.getfullargspec(callable)
        if len(qualname_splited) != 2:
            raise Exception(f'当前方法{callable.__qualname__}不是类方法')
        if (len(arg_spec.args) != 2 or arg_spec.varargs is not None
                or arg_spec.varkw is not None or arg_spec.defaults is not None
                or len(arg_spec.kwonlyargs) != 0 or arg_spec.kwonlydefaults is not None
                or arg_spec.annotations.get(arg_spec.args[1]) is None):
            raise Exception(f'请检查订阅方法，当前方法{callable.__qualname__}应为Class.Method(self, event: EventType)形式')


import abc
from abc import abstractmethod


class Logger(abc.ABC):

    @abstractmethod
    def trace(self, msg):
        pass

    @abstractmethod
    def debug(self, msg):
        pass

    @abstractmethod
    def info(self, msg):
        pass

    @abstractmethod
    def warning(self, msg):
        pass

    @abstractmethod
    def error(self, msg):
        pass

    @abstractmethod
    def exception(self, msg):
        pass

    @abstractmethod
    def critical(self, msg):
        pass



import abc
import queue
import threading


class PendingPost:

    def __init__(self, subscription, event):
        self.subscription = subscription
        self.event = event


class Executor(abc.ABC):

    @abc.abstractmethod
    def enqueue(self, subscription, event):
        pass


class MainExecutor(Executor):

    def __init__(self, eventbus):
        self.bus = eventbus

    def enqueue(self, subscription, event):
        self.bus.invoke_subscriber(subscription, event)


class MainOrderExecutor(Executor):

    def __init__(self, eventbus):
        self.quit = False
        self.queue = queue.Queue()
        self.bus = eventbus
        self.thread = threading.Thread(target=self.run)
        self.thread.start()

    def enqueue(self, subscription, event):
        pending_post = PendingPost(subscription, event)
        self.queue.put(pending_post)

    def run(self):
        while not self.quit:
            pending_post = self.queue.get()
            self.bus.invoke_subscriber(pending_post.subscription, pending_post.event)

    def stop(self):
        self.quit = True


import logging
import typing

from .executor import MainOrderExecutor, MainExecutor
from .finder import Finder
from .logger import Logger
from .subscription import Subscription


def get_default_logger(identifier):
    logger = logging.getLogger(f'EventBus[{identifier}]')
    logger.setLevel(logging.DEBUG)
    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(logging.WARNING)
    # create formatter
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    # add formatter to ch
    ch.setFormatter(formatter)
    # add ch to logger
    logger.addHandler(ch)
    return logger


class EventBus:

    def __init__(self, logger: Logger = None):
        self.finder = Finder()
        self.subscriptions_by_event_type: typing.Dict[type, typing.List[Subscription]] = {}
        self.event_types_by_subscriber: typing.Dict[object, typing.List[type]] = {}
        self.sticky_events: typing.Dict[type, object] = {}
        self.logger = logger if logger is not None else get_default_logger(id(self))
        self.executor = MainOrderExecutor(self)
        # self.executor = MainExecutor(self)

    def register(self, subscriber):
        if self.is_registered(subscriber):
            self.logger.warning(f"{subscriber}已经注册过")
            return
        subscriber_methods = self.finder.find(subscriber)
        for subscriber_method in subscriber_methods:
            self.subscribe(subscriber, subscriber_method)
            self.logger.debug(f'{subscriber} has registered the method {subscriber_method}')

    def is_registered(self, subscriber) -> bool:
        return subscriber in self.event_types_by_subscriber

    def subscribe(self, subscriber, subscriber_method):
        event_type = subscriber_method.event_type
        subscription = Subscription(subscriber, subscriber_method)
        subscriptions = self.subscriptions_by_event_type.get(event_type, [])
        if subscription in subscriptions:
            raise Exception(f'订阅类{subscriber.__class__.__name__}已经注册了{event_type.__name__}事件')
        subscriptions.append(subscription)
        # 订阅方法重排序
        subscriptions.sort(key=lambda _subscription: _subscription.subscriber_method.priority)
        self.subscriptions_by_event_type.update({event_type: subscriptions})
        # 按照订阅者绑定事件类型
        subscriber_events = self.event_types_by_subscriber.get(subscriber, [])
        subscriber_events.append(event_type)
        self.event_types_by_subscriber.update({subscriber: subscriber_events})

        if subscriber_method.sticky:
            sticky_event = self.sticky_events.get(event_type, None)
            self.check_post_sticky_event_to_subscription(subscription, sticky_event)

    def check_post_sticky_event_to_subscription(self, subscription: Subscription, event):
        if event is not None:
            self.invoke_subscriber(subscription, event)

    def unregister(self, subscriber):
        subscribed_types = self.event_types_by_subscriber.get(subscriber)
        if subscribed_types is None:
            self.logger.warning(f"订阅者{subscriber}没有被正常注册故无需注销")
        else:
            for event_type in subscribed_types:
                self.unsubscribe_by_event_type(subscriber, event_type)

    def unsubscribe_by_event_type(self, subscriber, event_type):
        """按照事件类型取消订阅

        仅更新subscriptions_by_event_type,不更新types_by_subscriber，
        所以在调用该方法时必须提前调用更新types_by_subscriber
        :param subscriber:
        :param event_type:
        :return:
        """
        # 获取事件对应的subscription列表
        subscriptions = self.subscriptions_by_event_type.get(event_type, None)
        if subscriptions is None:
            return
        # 获取subscription列表中包含subscriber的元素
        for subscription in subscriptions:
            if subscription.subscriber == subscriber:
                # 失活状态
                subscription.active = False
                # 删除
                subscriptions.remove(subscription)

    def post(self, event):
        self.logger.info(f"正在发布事件{event}")
        subscriptions = self.subscriptions_by_event_type.get(event.__class__, [])
        for subscription in subscriptions:
            self.executor.enqueue(subscription, event)

    def post_sticky(self, event):
        self.sticky_events.update({event.__class__: event})
        self.post(event)

    def get_sticky_event(self, event_type):
        return self.sticky_events.get(event_type, None)

    def remove_sticky_event(self, event):
        """
        支持对事件类、事件实例进行解析并移除
        :param event: 事件类或事件实例
        :return:
        """
        if isinstance(event, type):
            del self.sticky_events[event]
        elif self.sticky_events.get(event.__class__) == event:
            del self.sticky_events[event.__class__]

    def cancel_delivery(self, event):
        setattr(event, '__cancelled__', True)
        if self.sticky_events.get(event.__class__) == event:
            self.remove_sticky_event(event)

    def invoke_subscriber(self, subscription, event):
        if subscription.active and not getattr(event, '__cancelled__', False):
            try:
                subscription.subscriber_method(subscription.subscriber, event)
            except Exception as exception:
                self.on_exception(subscription, event, exception)

    def on_exception(self, subscription, event, exception):
        self.logger.warning(f'在为{subscription}发送{event}时出现错误:{exception}')

    def stop(self):
        self.executor.stop()





class Subscription:

    def __init__(self, subscriber, subscriber_method):
        self.subscriber = subscriber
        self.subscriber_method = subscriber_method
        self.active = True

    def __eq__(self, other):
        if isinstance(other, Subscription):
            if other.subscriber == self.subscriber and other.subscriber_method == self.subscriber_method:
                return True
        return False

    def __str__(self):
        return f'<Subscription[{self.subscriber.__class__.__name__}@{hex(id(self.subscriber))} # {self.subscriber_method.method.__name__}@{hex(id(self.subscriber_method))}] at {hex(id(self))}>'


import inspect

from eventbus import subscribe
from eventbus.subscriber_method import SubscriberMethod


class Finder:

    def find(self, subscriber):
        return self._find_by_reflect(subscriber)

    def _find_by_reflect(self, subscriber):
        subscriber_methods = []

        methods = inspect.getmembers(subscriber, predicate=inspect.ismethod)
        for name, method in methods:
            annotation = getattr(method, subscribe.subscribe_annotation_property, None)
            if annotation is None:
                continue
            assert len(method.__annotations__) == 1, '订阅方法参数不合法'
            event_type = list(method.__annotations__.values())[0]
            subscriber_method = SubscriberMethod(method.__func__, event_type, priority=annotation.priority, sticky=annotation.sticky)
            subscriber_methods.append(subscriber_method)

        return subscriber_methods
